import logging
from typing import Optional
import fsspec

import pandas as pd
import xarray as xr

from ufs2arco.sources import EnsembleForecastSource

logger = logging.getLogger("ufs2arco")

class AWSGEFSArchive(EnsembleForecastSource):
    """Access NOAA's forecast archive from the Global Ensemble Forecast System (GEFS) at s3://noaa-gefs-pds."""

    static_vars = ("lsm", "orog")
    sample_dims = ("t0", "fhr", "member")
    base_dims = ("latitude", "longitude")

    @property
    def available_variables(self) -> tuple:
        return tuple(self._ic_variables.keys())

    @property
    def available_levels(self) -> tuple:
        return (
            10, 20, 30, 50, 70,
            100, 150, 200, 250, 300, 350, 400, 450,
            500, 550, 600, 650, 700, 750, 800, 850,
            900, 925, 950, 975, 1000,
        )

    @property
    def rename(self) -> dict:
        return {
            "time": "t0",
            "step": "lead_time",
            "isobaricInhPa": "level",
        }

    def open_sample_dataset(
        self,
        t0: pd.Timestamp,
        fhr: int,
        member: int,
        open_static_vars: bool,
        cache_dir: Optional[str] = None,
    ) -> xr.Dataset:

        # 1. cache the grib files for this date, member, fhr
        cached_files = {}
        for k in ["a", "b"]:
            path = self._build_path(
                t0=t0,
                member=member,
                fhr=fhr,
                a_or_b=k,
            )
            if cache_dir is not None:
                try:
                    local_file = fsspec.open_local(
                        path,
                        s3={"anon": True},
                        filecache={"cache_storage": cache_dir},
                    )
                except FileNotFoundError:
                    local_file = None
                    logger.warning(
                        f"{self.name}: File Not Found: {path}\n\t" +
                        f"(t0, member, fhr, key) = {t0} {member} {fhr} {k}"
                    )
                except:
                    local_file = None
                    logger.warning(
                        f"{self.name}: Trouble finding the file: {path}\n\t" +
                        f"(t0, member, fhr, key) = {t0} {member} {fhr} {k}"
                    )
                cached_files[k] = local_file
            else:
                cached_files[k] = path

        # 2. read data arrays from those files
        dsdict = {}
        if cached_files["a"] is not None and cached_files["b"] is not None:
            read_dict = self._ic_variables if open_static_vars else self._fc_variables
            read_dict = {k: v for k,v in read_dict.items() if k in self.variables}
            for varname, open_kwargs in read_dict.items():
                dslist = []
                for a_or_b in open_kwargs["param_list"]:
                    try:
                        thisvar = self._open_single_variable(
                            file=cached_files[a_or_b],
                            varname=varname,
                            member=member,
                            filter_by_keys=open_kwargs["filter_by_keys"],
                        )
                    except:
                        thisvar = None
                        logger.warning(
                            f"{self.name}: Trouble opening {varname}\n\t" +
                            f"(t0, member, fhr, key) = {t0} {member} {fhr} {a_or_b}"
                        )
                    dslist.append(thisvar)
                if not any(x is None for x in dslist):
                    dsdict[varname] = xr.merge(dslist)[varname]
                else:
                    dsdict[varname] = xr.DataArray(name=varname)
        xds = xr.Dataset(dsdict)
        return xds

    def _open_single_variable(
        self,
        file: fsspec.spec.AbstractFileSystem,
        varname: str,
        member: int,
        filter_by_keys: dict,
    ) -> xr.DataArray:
        """
        Open a single variable from a GRIB file.

        Args:
            file (fsspec.spec.AbstractFileSystem): The file to read.
            varname (str): The variable name to extract.
            member (int): The ensemble member ID.
            filter_by_keys (dict): Keys to filter the variable by.

        Returns:
            xr.DataArray: The extracted variable as an xarray DataArray.
        """
        xds = xr.open_dataset(file, engine="cfgrib", filter_by_keys=filter_by_keys)
        xda = xds[varname]

        if "isobaricInhPa" in xds.coords:
            if len(xda.dims) < 3:
                vv = xda["isobaricInhPa"].values
                xda = xda.expand_dims({"isobaricInhPa": [vv]})

        for v in ["heightAboveGround", "number", "surface"]:
            if v in xda.coords:
                xda = xda.drop_vars(v)


        xds = xda.to_dataset(name=varname)
        for key, val in self.rename.items():
            if key in xds:
                xds = xds.rename({key: val})
        if varname in self.static_vars:
            for key in ["lead_time", "t0", "valid_time"]:
                if key in xds:
                    xds = xds.drop_vars(key)
        else:

            if "level" in xds:
                level_selection = [l for l in self.levels if l in xds.level.values]
                if len(level_selection) == 0:
                    return xr.DataArray(name=varname)
                xds = xds.sel(level=level_selection, **self._level_sel_kwargs)

            xds = xds.expand_dims(["t0", "lead_time", "member"])
            xds["fhr"] = xr.DataArray(
                int(xds["lead_time"].values / 1e9 / 3600),
                coords=xds["lead_time"].coords,
                attrs={
                    "long_name": "hours since initial time",
                    "units": "integer hours",
                },
            )
            xds["member"] = xr.DataArray(
                [member],
                coords={"member": [member]},
                dims=("member",),
                attrs={
                    "description": "ID=0 comes from gecXX files, ID>0 comes from gepXX files",
                    "long_name": "ensemble member ID",
                },
            )

            # recreate valid_time, since it's not always there
            xds = xds.swap_dims({"lead_time": "fhr"})
            valid_time = xds["t0"] + xds["lead_time"]
            if "valid_time" in xds:
                xds["valid_time"] = xds["valid_time"].expand_dims(["t0", "fhr"])
                assert valid_time.squeeze() == xds.valid_time.squeeze()
                xds = xds.drop_vars("valid_time")

            xds["valid_time"] = valid_time
            xds = xds.set_coords("valid_time")
        return xds[varname]


    def _build_path(self, t0: pd.Timestamp, member: int, fhr: int, a_or_b: str) -> str:
        """
        Build the file path to a GRIB file based on the provided parameters.

        Args:
            t0 (pd.Timestamp): The initial condition timestamp.
            member (int): The ensemble member ID.
            fhr (int): The forecast hour.
            a_or_b (str): The file type, either 'a' or 'b'.

        Returns:
            str: The constructed file path.
        """
        c_or_p = "c" if member == 0 else "p"
        bucket = f"s3://noaa-gefs-pds"
        outer = f"gefs.{t0.year:04d}{t0.month:02d}{t0.day:02d}/{t0.hour:02d}"
        # Thanks to Herbie for figuring these out
        if t0 < pd.Timestamp("2018-07-27"):
            middle = ""
            fname = f"ge{c_or_p}{member:02d}.t{t0.hour:02d}z.pgrb2{a_or_b}f{fhr:03d}"
        elif t0 < pd.Timestamp("2020-09-23T12"):
            middle = f"pgrb2{a_or_b}/"
            fname = f"ge{c_or_p}{member:02d}.t{t0.hour:02d}z.pgrb2{a_or_b}f{fhr:02d}"
        else:
            middle = f"atmos/pgrb2{a_or_b}p5/"
            fname = f"ge{c_or_p}{member:02d}.t{t0.hour:02d}z.pgrb2{a_or_b}.0p50.f{fhr:03d}"
        fullpath = f"filecache::{bucket}/{outer}/{middle}{fname}"
        logger.debug(f"{self.name}._build_path: reading {fullpath}")
        return fullpath

    @property
    def _ic_variables(self) -> dict:
        """
        Get the dictionary of initial condition variables.

        Returns:
            dict: The dictionary of variables for initial conditions.
        """
        return {
            "lsm": {
                "param_list": ["b"],
                "filter_by_keys": {
                    "typeOfLevel": "surface",
                    "paramId": [172],
                },
            },
            "orog": {
                "param_list": ["a"],
                "filter_by_keys": {
                    "typeOfLevel": "surface",
                    "paramId": [228002],
                },
            },
            "sp": {
                "param_list": ["a"],
                "filter_by_keys": {
                    "typeOfLevel": "surface",
                    "paramId": [134],
                },
            },
            "u10": {
                "param_list": ["a"],
                "filter_by_keys": {
                    "typeOfLevel": "heightAboveGround",
                    "paramId": [165],
                },
            },
            "v10": {
                "param_list": ["a"],
                "filter_by_keys": {
                    "typeOfLevel": "heightAboveGround",
                    "paramId": [166],
                },
            },
            "t2m": {
                "param_list": ["a"],
                "filter_by_keys": {
                    "typeOfLevel": "heightAboveGround",
                    "paramId": [167],
                },
            },
            "sh2": {
                "param_list": ["b"],
                "filter_by_keys": {
                    "typeOfLevel": "heightAboveGround",
                    "paramId": [174096],
                },
            },
            "gh": {
                "param_list": ["a", "b"],
                "filter_by_keys": {
                    "typeOfLevel": "isobaricInhPa",
                    "paramId": [156],
                },
            },
            "u": {
                "param_list": ["a", "b"],
                "filter_by_keys": {
                    "typeOfLevel": "isobaricInhPa",
                    "paramId": [131],
                },
            },
            "v": {
                "param_list": ["a", "b"],
                "filter_by_keys": {
                    "typeOfLevel": "isobaricInhPa",
                    "paramId": [132],
                },
            },
            "w": {
                "param_list": ["a", "b"],
                "filter_by_keys": {
                    "typeOfLevel": "isobaricInhPa",
                    "paramId": [135],
                },
            },
            "t": {
                "param_list": ["a", "b"],
                "filter_by_keys": {
                    "typeOfLevel": "isobaricInhPa",
                    "paramId": [130],
                },
            },
            "q": {
                "param_list": ["b"],
                "filter_by_keys": {
                    "typeOfLevel": "isobaricInhPa",
                    "paramId": [133],
                },
            },
        }

    @property
    def _fc_variables(self):
        """
        Get the dictionary of forecast variables, which is notably different from
        the initial condition files.

        Returns:
            dict: The dictionary of variables for forecast files/variables
        """
        fckw = self._ic_variables.copy()
        for key in self.static_vars:
            fckw.pop(key)
        return fckw



import os
import shutil
import itertools
import logging

from math import ceil

import xarray as xr
import dask.array

from .mpi import MPITopology, _has_mpi

logger = logging.getLogger("ufs2arco")

class DataMover():
    """Move data, using the concept of a data "sample" to define how much data is stored to zarr at once.
    A data sample is defined by :attr:`sample_dims`. For example, if a dataset has dimensions
    ``("t0", "fhr", "member", "pressure", "latitude", "longitude")``
    then setting ``sample_dims = ("t0", "fhr", "member")``
    would loop through the data using those indices, where each sample has the full
    ``("pressure", "latitude", "longitude")`` array for the given values of ``("t0", "fhr", "member")``.

    Note:
        * For now only works with GEFSDataset
        * This is the same as loop through the data storing one sample at a time,
          except that it stores ``batch_size`` samples in a hard disk cache.
          It clears the cache after every batch.
        * Multithreading for simultaneous reading/writing could be implemented, but hasn't seemed necessary.
          Also it would double (or more) the cache requirements.
    """
    counter = 0
    data_counter = 0

    stop_event = None
    executor = None
    futures = None

    def __init__(
        self,
        source,
        target,
        batch_size,
        transformer=None,
        start=0,
        cache_dir=".",
    ):

        self.source = source
        self.target = target
        self.transformer = transformer if transformer is not None else lambda xds: xds
        self.batch_size = batch_size
        self.counter = start
        self.data_counter = start
        self.outer_cache_dir = cache_dir

        # construct the sample indices
        # e.g. {"t0": [date1, date2], "fhrs": [0, 6], "member": [0, 1, 2]}
        all_sample_iterations = {
            key: getattr(source, key)
            for key in self.source.sample_dims
        }
        self.sample_indices = list()
        for combo in itertools.product(*all_sample_iterations.values()):
            self.sample_indices.append(dict(zip(self.source.sample_dims, combo)))

        self.restart(idx=start)

    @property
    def name(self):
        return str(type(self).__name__)


    def __len__(self) -> int:
        n_samples = len(self.sample_indices)
        n_batches = ceil(n_samples / self.batch_size)
        return n_batches

    def __iter__(self):
        self.counter = 0
        self.restart()
        return self

    def __next__(self):
        if self.counter < len(self):
            data = self.get_data()
            self.counter += 1
            return data
        else:
            logger.debug(f"{self.name}.__next__: counter > len(self)")
            raise StopIteration


    def get_cache_dir(self, batch_idx):
        return f"{self.outer_cache_dir}/{self.name.lower()}-cache/{batch_idx}"


    def get_batch_indices(self, batch_idx):
        st = batch_idx * self.batch_size
        ed = st + self.batch_size
        return self.sample_indices[st:ed]


    def _next_data(self):
        logger.debug(f"{self.name}._next_data[{self.data_counter}]")
        if self.data_counter < len(self):
            batch_indices = self.get_batch_indices(self.data_counter)
            cache_dir = self.get_cache_dir(self.data_counter)
            if len(batch_indices) > 0:
                dlist = []

                for these_dims in batch_indices:
                    fds = self.source.open_sample_dataset(
                        open_static_vars=self.target.always_open_static_vars,
                        cache_dir=cache_dir,
                        **these_dims,
                    )
                    fds = self.transformer(fds)
                    fds = self.target.apply_transforms_to_sample(fds)
                    dlist.append(fds)
                xds = xr.merge(dlist)
                return xds

            else:
                return None
        else:
            logger.debug(f"{self.name}._next_data: data_counter > len(self)")
            raise StopIteration


    def get_data(self):
        """Pull a batch of data from the queue"""
        logger.debug(f"{self.name}.get_data")
        data = self._next_data()
        self.data_counter += 1
        return data


    def clear_cache(self, batch_idx):
        cache_dir = self.get_cache_dir(batch_idx)
        if os.path.isdir(cache_dir):
            logger.debug(f"{self.name}: clearing {cache_dir}")
            shutil.rmtree(cache_dir, ignore_errors=True)

    def restart(self, idx=0):
        """Restart the :attr:`data_counter` to get ready for the pass through the data

        Args:
            idx (int, optional): index to restart to
        """
        logger.debug(f"{self.name}.restart: idx = {idx}")
        self.data_counter = idx


    def create_container(self, **kwargs) -> None:
        """
        Create a Zarr container to store the dataset.

        Args:
            **kwargs: Additional arguments passed to `xr.Dataset.to_zarr`.

        Logs:
            The created container is stored at `self.target.store_path`.
        """

        # open a minimal dataset
        sample_dim_args = {
            key: getattr(self.source, key)[0]
            for key in self.source.sample_dims
        }
        xds = self.source.open_sample_dataset(
            **sample_dim_args,
            open_static_vars=True,
            cache_dir=self.get_cache_dir("container"),
        )

        # perform any transformations like regridding
        xds = self.transformer(xds)

        # transform it to target space
        xds = self.target.apply_transforms_to_sample(xds)

        # create container
        # start with the dimensions we haven't read yet (sample_dims)
        nds = xr.Dataset(attrs=xds.attrs.copy())
        for key in self.target.renamed_sample_dims:
            array = getattr(self.target, key)
            nds[key] = xr.DataArray(
                array,
                coords={key: array},
                dims=key,
                attrs=xds[key].attrs.copy(),
            )

        # these will be the base_dims, we read these in each sample
        for key in xds.dims:
            if key not in self.target.renamed_sample_dims:
                nds[key] = xds[key].copy()

        # manage coordinates
        # first we have to pass all existing coordinates,
        # then let target manage them
        for key in xds.coords:
            if key not in nds:
                nds = nds.assign_coords({key: xds[key].copy()})
        nds = self.target.manage_coords(nds)

        # create empty data arrays
        for varname in xds.data_vars:
            dims = xds[varname].dims
            shape = tuple(len(nds[key]) for key in dims)
            chunks = {list(dims).index(key): self.target.chunks[key] for key in dims}
            nds[varname] = xr.DataArray(
                data=dask.array.zeros(
                    shape=shape,
                    chunks=chunks,
                    dtype=xds[varname].dtype,
                ),
                dims=dims,
                attrs=xds[varname].attrs.copy(),
            )

        nds.to_zarr(self.target.store_path, compute=False, **kwargs)
        logger.info(f"{self.name}.create_container: stored container at {self.target.store_path}\n{nds}\n")

        self.clear_cache("container")


    def find_my_region(self, xds):
        """Given a dataset, that's assumed to be a subset of the initial dataset,
        find the logical index values where this should be stored in the final zarr store

        Args:
            xds (xr.Dataset): with a subset of the data (i.e., a couple of initial conditions)

        Returns:
            region (dict): indicating the zarr region to store in, based on the initial condition indices
        """
        region = {k: slice(None, None) for k in xds.dims}
        for key in self.target.renamed_sample_dims:
            full_array = getattr(self.target, key) # e.g. all of the initial conditions
            batch_indices = [list(full_array).index(value) for value in xds[key].values]
            region[key] = slice(batch_indices[0], batch_indices[-1]+1)
        return region



class MPIDataMover(DataMover):
    """Use MPI to scale up the data moving. Note that this pulls in one data sample per MPI proces
    (i.e., :attr:`data_per_process` always = 1), and uses the MPI size (i.e., number of processes)
    as the :attr:`batch_size`.

    So, this also requires ``num_mpi_processes`` = :attr:`batch_size` * `sample_size_in_bytes` disk space
    for caching.
    """
    def __init__(
        self,
        source,
        target,
        mpi_topo,
        transformer=None,
        start=0,
        cache_dir=".",
    ):
        assert _has_mpi, f"{self.name}.__init__: Unable to import mpi4py, cannot use this class"

        self.topo = mpi_topo
        batch_size = self.topo.size
        self.data_per_process = 1
        self.local_batch_index = self.topo.rank*self.data_per_process
        super().__init__(
            source=source,
            target=target,
            batch_size=batch_size,
            transformer=transformer,
            start=start,
            cache_dir=cache_dir,
        )
        logger.info(str(self))

    def __str__(self):
        title = f"Mover: {self.name}"
        underline = "".join(["-" for _ in range(len(title))])
        msg = f"\n{title}\n{underline}\n"

        for key in ["local_batch_index", "data_per_process", "batch_size"]:
            msg += f"{key:<18s}: {getattr(self, key):02d}\n"

        msg += f"{'Total Samples':<18s}: {len(self.sample_indices)}\n"
        msg += f"{'Total Batches':<18s}: {len(self)}\n"
        msg += f"{'sample_dims':<18s}: {self.source.sample_dims}\n"
        return msg

    def get_cache_dir(self, batch_idx):
        return f"{self.outer_cache_dir}/{self.name.lower()}-cache/{self.topo.rank}/{batch_idx}"

    def get_batch_indices(self, batch_idx):
        st = (batch_idx * self.batch_size) + self.local_batch_index
        ed = st + self.data_per_process
        return self.sample_indices[st:ed]

    def create_container(self, **kwargs) -> None:
        if self.topo.is_root:
            super().create_container(**kwargs)
        self.topo.barrier()



import os
import yaml
from typing import Dict, List, Callable
from datetime import datetime
import numpy as np
import xarray as xr

from .ufsdataset import UFSDataset

class FV3Dataset(UFSDataset):
    __doc__ = UFSDataset.__doc__
    def __init__(self, path_in: Callable, config_filename: str, is_nested: bool = False) -> None:
        super(FV3Dataset, self).__init__(path_in, config_filename, is_nested)
        self.zarr_name = "fv3.zarr"
        self.chunks_in = self.chunks_in if len(self.chunks_in) != 0 else {
            "pfull": 1,
            "grid_yt": -1,
            "grid_xt": -1,
        }

        self.chunks_out = self.chunks_out if len(self.chunks_out) != 0 else {
            "time": 1,
            "pfull": 1,
            "grid_yt": -1,
            "grid_xt": -1,
        }

    def open_dataset(self, cycles: datetime, fsspec_kwargs=None, **kwargs):
        xds = super().open_dataset(cycles, fsspec_kwargs, **kwargs)

        # Deal with time
        xds = xds.rename({"time": "cftime"})
        xds["time"] = self._cftime2time(xds["cftime"])
        xds["ftime"] = self._time2ftime(xds["time"], cycles)
        xds = xds.swap_dims({"cftime": "time"})
        xds = xds.set_coords(["time", "cftime", "ftime"])

        # convert ak/bk attrs to coordinate arrays
        for key in ["ak", "bk"]:
            if key in xds.attrs:
                xds[key] = xr.DataArray(
                    xds.attrs.pop(key),
                    coords=xds["phalf"].coords,
                    dims=xds["phalf"].dims,
                )
                xds = xds.set_coords(key)

        # rename grid_yt.long_name to avoid typo
        xds["grid_yt"].attrs["long_name"] = "T-cell latitude"
        return xds





import logging
from typing import Optional
import yaml

import pandas as pd
import xarray as xr

from ufs2arco.sources import AnalysisSource

logger = logging.getLogger("ufs2arco")


class GCSReplayAtmosphere(AnalysisSource):
    """
    Atmospheric component of replay, already zarr-ified on GCS.

    These zarr stores could very easily be generalized... I'm sure.
    """

    static_vars = ("land_static", "hgtsfc_static")

    _is_selected = False

    @property
    def rename(self) -> dict:
        return {
            "pfull": "level",
            "grid_yt": "latitude",
            "grid_xt": "longitude",
        }

    @property
    def available_variables(self) -> tuple:
        return tuple(self._xds.data_vars)

    @property
    def available_levels(self) -> tuple:
        return tuple(self._xds["level"].values)

    def __init__(
        self,
        uri: str,
        time: dict,
        variables: Optional[list | tuple] = None,
        levels: Optional[list | tuple] = None,
        use_nearest_levels: Optional[bool] = False,
    ) -> None:

        # open and rename
        xds = xr.open_zarr(
            uri,
            storage_options={"token": "anon"},
        )
        self._xds = xds.rename(self.rename)

        # parent class checks if variables and levels are legit
        super().__init__(
            time=time,
            variables=variables,
            levels=levels,
            use_nearest_levels=use_nearest_levels,
        )

        # now subsample the dataset
        self._xds = self._xds[self.variables]
        self._xds = self._xds.sel(level=self.levels, **self._level_sel_kwargs)

        # drop these because cftime gives trouble no matter what
        self._xds = self._xds.drop_vars(["cftime", "ftime"])

    def open_sample_dataset(
        self,
        time: pd.Timestamp,
        open_static_vars: bool,
        cache_dir: Optional[str] = None,
    ) -> xr.Dataset:

        xds = self._xds.sel(time=[time])
        selection = list(self.variables) if open_static_vars else list(self.dynamic_vars)
        xds = xds[selection].copy().load()
        return xds



import os
import yaml
import warnings
import xarray as xr
import numpy as np
import cf_xarray as cfxr

from .ufsregridder import UFSRegridder

try:
    import xesmf as xe
except ImportError:
    pass


class MOM6Regridder(UFSRegridder):
    """
    Regrid ocean dataset that is on a tripolar grid to a different grid (primarily Gaussian grid).


    Optional fields in config:
        rotation_file (str): path to file containing rotation fields "sin_rot" and "cos_rot", required for regridding vector fields
        weights_file_t2t (str): path to t2t interpolation weights file
        weights_file_u2t (str): path to u2t interpolation weights file
        weights_file_v2t (str): path to v2t interpolation weights file
        periodic (bool): Is the grid periodic in longitude?
    """

    rg_tt = None
    rg_ut = None
    rg_vt = None

    __doc__ = __doc__ + UFSRegridder.__doc__

    def __init__(
        self,
        lats1d_out: np.array,
        lons1d_out: np.array,
        ds_in: xr.Dataset,
        config_filename: str,
    ) -> None:
        super(MOM6Regridder, self).__init__(
            lats1d_out, lons1d_out, ds_in, config_filename
        )

    def create_grid_in(
        self,
        mom6_grid: xr.Dataset,
    ) -> xr.Dataset:
        """Convert mom6 grid into a grid that is ready to be used by regridder.
        This comes from here: https://mom6-analysiscookbook.readthedocs.io/en/latest/notebooks/Horizontal_Remapping.html

        Args:
            mom6_grid (xr.Dataset):
                Dataset with all necessary mom6 metadata.

        Returns:
            ds_in_t (xr.Dataset):
                Dataset ready to be used as input grid for tracers.
            ds_in_v (xr.Dataset):
                Dataset ready to be used as input grid for v velocity.
            ds_in_u (xr.Dataset):
                Dataset ready to be used as input grid for u velocity.
        """
        grid_in = xr.Dataset()
        grid_in["lon"] = mom6_grid["geolon"]
        grid_in["lat"] = mom6_grid["geolat"]
        grid_in["lon_u"] = mom6_grid["geolon_u"]
        grid_in["lat_u"] = mom6_grid["geolat_u"]
        grid_in["lon_v"] = mom6_grid["geolon_v"]
        grid_in["lat_v"] = mom6_grid["geolat_v"]
        grid_in['cos_rot'] = mom6_grid["cos_rot"]
        grid_in['sin_rot'] = mom6_grid["sin_rot"]
        ny, nx = grid_in["lon"].shape
        lon_b = np.empty((ny + 1, nx + 1))
        lat_b = np.empty((ny + 1, nx + 1))
        lon_b[1:, 1:] = mom6_grid["geolon_c"].values
        lat_b[1:, 1:] = mom6_grid["geolat_c"].values
        # periodicity
        lon_b[:, 0] = lon_b[:, -1]
        lat_b[:, 0] = lat_b[:, -1]
        # south edge
        dy = (lat_b[2, :] - lat_b[1, :]).mean()
        lat_b[0, 1:] = lat_b[1, 1:] - dy
        lon_b[0, 1:] = lon_b[1, 1:]
        # corner point
        lon_b[0, 0] = lon_b[1, 0]
        lat_b[0, 0] = lat_b[0, 1]
        grid_in["lon_b"] = xr.DataArray(data=lon_b)
        grid_in["lat_b"] = xr.DataArray(data=lat_b)

        # create renamed datasets
        ds_in_t = grid_in[["lon", "lat", "lat_b", "lon_b"]]
        ds_in_u = grid_in[["lon_u", "lat_u", "lat_b", "lon_b"]].rename(
            {"lat_u": "lat", "lon_u": "lon"}
        )
        ds_in_v = grid_in[["lon_v", "lat_v", "lat_b", "lon_b"]].rename(
            {"lat_v": "lat", "lon_v": "lon"}
        )
        ds_rot =  grid_in[['cos_rot','sin_rot']]

        return ds_in_t, ds_in_u, ds_in_v, ds_rot
    
    def create_grid_out(
        self,
        lats: np.array,
        lons: np.array,
    ) -> xr.Dataset:
        """Take lat/lon of our grid and create a grid that is ready for regridder.

        Args:
            lats (np.array):
                Lats of grid_out.
            lons (xr.Dataset):
                Lons of grid_out.

        Returns:
            grid_out (xr.Dataset):
                Grid out that is ready for regridding.
        """
        grid_out = xr.Dataset()
        grid_out["lon"] = xr.DataArray(lons, dims=["lon"])
        grid_out["lat"] = xr.DataArray(lats, dims=["lat"])
        grid_out = grid_out.cf.add_bounds(["lat", "lon"])
        lat_corners = cfxr.bounds_to_vertices(
            bounds=grid_out["lat_bounds"], bounds_dim="bounds", order=None
        )
        lon_corners = cfxr.bounds_to_vertices(
            bounds=grid_out["lon_bounds"], bounds_dim="bounds", order=None
        )
        grid_out = grid_out.assign({"lat_b": lat_corners, "lon_b": lon_corners})
        grid_out = grid_out.drop_vars(["lat_bounds", "lon_bounds"])

        return grid_out
    
    def _create_regridder(self, ds_in: xr.Dataset) -> None:

        # create rotation dataset
        self.rotation_file = self.config.get("rotation_file", None)
        self.ds_rot = None
        if self.rotation_file is not None:
            mom6_grid = xr.open_dataset(self.rotation_file)
        else:
            warnings.warn(
                "MOM6Regridder._create_regridder: Could not find 'rotation_file' in configuration yaml. "
                "If using conservative regridding, it will fail. Otherwise, just vector fields will be ignored."
            )

        # create renamed datasets
        ds_in_t, ds_in_u, ds_in_v, self.ds_rot = self.create_grid_in(mom6_grid=mom6_grid)

        # create output dataset
        grid_out = self.create_grid_out(
            lats = self.lats1d_out,
            lons = self.lons1d_out,
        )

        # get nlon/nlat for input/output datsets
        nlon_i = ds_in.sizes["yh"]
        nlat_i = ds_in.sizes["xh"]
        nlon_o = len(self.lons1d_out)
        nlat_o = len(self.lats1d_out)
        self.ires = f"{nlon_i}x{nlat_i}"
        self.ores = f"{nlon_o}x{nlat_o}"

        # paths to interpolation weights files
        interp_method = self.config["interp_method"]
        wfiles = dict()
        for key in ["weights_file_t2t", "weights_file_u2t", "weights_file_v2t"]:
            vin = key[-3]
            default = f"weights-mom6-{self.ires}.C{vin}.{self.ores}.Ct.{interp_method}.nc"
            path = self.config.get(key, None)
            wfiles[key] = path if path is not None else default

        # create regridding instances
        periodic = self.config["periodic"]
        reuse = os.path.exists(wfiles["weights_file_t2t"])
        self.rg_tt = xe.Regridder(
            ds_in_t,
            grid_out,
            interp_method,
            periodic=periodic,
            reuse_weights=reuse,
            filename=wfiles["weights_file_t2t"],
            unmapped_to_nan=True,
        )
        if self.ds_rot is not None:
            reuse = os.path.exists(wfiles["weights_file_u2t"])
            self.rg_ut = xe.Regridder(
                ds_in_u,
                ds_in_t,
                interp_method,
                periodic=periodic,
                reuse_weights=reuse,
                filename=wfiles["weights_file_u2t"],
                unmapped_to_nan=True,
            )
            reuse = os.path.exists(wfiles["weights_file_v2t"])
            self.rg_vt = xe.Regridder(
                ds_in_v,
                ds_in_t,
                interp_method,
                periodic=periodic,
                reuse_weights=reuse,
                filename=wfiles["weights_file_v2t"],
                unmapped_to_nan=True,
            )

    def regrid(self, ds_in: xr.Dataset) -> xr.Dataset:

        # define MOM6 dataset specific coordinates and vector fields map
        coords_xy = {"yh", "xh", "yq", "xq"}
        variable_map = {
            "SSU": ("SSV", "U"),
            "SSV": (None, "skip"),
            "uo": ("vo", "U"),
            "vo": (None, "skip"),
            "taux": ("tauy", "U"),
            "tauy": (None, "skip"),
        }

        return super(MOM6Regridder, self).regrid_tripolar(
            ds_in,
            self.ds_rot,
            self.rg_tt,
            self.rg_ut,
            self.rg_vt,
            coords_xy,
            variable_map,
        )



import logging
from typing import Optional
import numpy as np
import pandas as pd
import xarray as xr

from ufs2arco.sources import Source

logger = logging.getLogger("ufs2arco")

class EnsembleForecastSource(Source):
    """
    Base class for all ensemble forecast-like datasets.

    Child classes must map dimension names to:
        ``("time", "member", "level", "latitude", "longitude")``
    """

    sample_dims = ("t0", "fhr", "member")
    base_dims = ("level", "latitude", "longitude")

    # fill these out per subclass
    static_vars = tuple()
    available_variables = tuple()
    available_levels = tuple()

    @property
    def rename(self) -> dict:
        """
        Use this to map to the dimensions
        (t0, fhr, member, level, latitude, longitude)
        """
        return dict()

    def __init__(
        self,
        t0: dict,
        fhr: dict,
        member: dict,
        variables: Optional[list | tuple] = None,
        levels: Optional[list | tuple] = None,
        use_nearest_levels: Optional[bool] = False,
    ) -> None:
        """
        Initialize the Source object.

        Args:
            t0 (dict): Dictionary with start and end times for initial conditions, and e.g. "freq=6h". All options get passed to ``pandas.date_range``.
            fhr (dict): Dictionary with 'start', 'end', and 'step' forecast hours.
            member (dict): Dictionary with 'start', 'end', and 'step' ensemble members.
            variables (list, tuple, optional): variables to grab
            levels (list, tuple, optional): vertical levels to grab
            use_nearest_levels (bool, optional): if True, all level selection with
                ``xarray.Dataset.sel(level=levels, method="nearest")``
        """
        self.t0 = pd.date_range(**t0)
        self.fhr = np.arange(fhr["start"], fhr["end"] + 1, fhr["step"])
        self.member = np.arange(member["start"], member["end"] + 1, member["step"])
        super().__init__(
            variables=variables,
            levels=levels,
            use_nearest_levels=use_nearest_levels,
        )


    def open_sample_dataset(
        self,
        t0: pd.Timestamp,
        fhr: int,
        member: int,
        open_static_vars: bool,
        cache_dir: Optional[str]=None,
    ) -> xr.Dataset:
        """
        Open a single dataset for the given initial condition, forecast hour, and member.

        The resulting dataset should always have dimensions:

        (t0, fhr, member, level, latitude, longitude)

        Even if any one of those is single

        Args:
            t0 (pd.Timestamp): The initial condition timestamp.
            fhr (int): The forecast hour.
            member (int): The ensemble member ID.
            open_static_vars (bool): if True, read static_vars, otherwise don't
            cache_dir (str, optional): Directory to cache files locally.

        Returns:
            xr.Dataset: The dataset containing the specified data.
        """
        pass


    def add_full_extra_coords(self, xds: xr.Dataset) -> xr.Dataset:
        # compute the full version of these extra coords
        xds["lead_time"] = xr.DataArray(
            [pd.Timedelta(hours=x) for x in self.fhr],
            coords=xds["fhr"].coords,
            attrs=xds["lead_time"].attrs.copy(),
        )

        xds["valid_time"] = xds["t0"] + xds["lead_time"]
        xds["valid_time"].attrs = xds["valid_time"].attrs.copy()
        xds = xds.set_coords(["lead_time", "valid_time"])
        return xds



import pytest

from os.path import join, dirname

from ufs2arco import FV3Dataset


@pytest.mark.parametrize(
    "prefix", ["gcs://", "s3://", "https://", "/contrib", "/scratch/"]
)
def test_join_cloud(prefix):

    dummy_path = lambda p : str(p)
    fname = join(dirname(__file__), "config-replay.yaml")
    ufs = FV3Dataset(dummy_path, fname)

    path = ufs._join(prefix, "directory", "fv3.zarr")
    if "://" in prefix:
        assert path == f"{prefix}directory/fv3.zarr"
    else:
        assert path == join(prefix, "directory", "fv3.zarr")



import pytest
from unittest.mock import patch, MagicMock
import pandas as pd
import numpy as np
import xarray as xr

from ufs2arco.sources import AWSGEFSArchive

@pytest.fixture
def gefs_dataset():
    t0 = {"start": "2017-01-01T00", "end": "2017-01-01T18", "freq": "6h"}
    fhr = {"start": 0, "end": 6, "step": 6}
    member = {"start": 0, "end": 1, "step": 1}
    chunks = {"t0": 1, "fhr": 1, "member": 1, "latitude": -1, "longitude": -1}
    return AWSGEFSArchive(t0, fhr, member)

def test_init(gefs_dataset):
    assert len(gefs_dataset.t0) == 4
    assert np.array_equal(gefs_dataset.fhr, np.array([0, 6]))
    assert np.array_equal(gefs_dataset.member, np.array([0, 1]))

def test_str(gefs_dataset):
    str(gefs_dataset) # just make sure this can run without bugs

def test_name(gefs_dataset):
    assert gefs_dataset.name == "AWSGEFSArchive"

def test_open_sample_dataset(gefs_dataset):
    result = gefs_dataset.open_sample_dataset(pd.Timestamp("2017-01-01T00"), 0, 0, "/tmp/cache", True)
    assert isinstance(result, xr.Dataset)



from .base import Source
from .analysis import AnalysisSource
from .ensemble_forecast import EnsembleForecastSource

from .aws_gefs_archive import AWSGEFSArchive
from .gcs_replay_atmosphere import GCSReplayAtmosphere
from .gcs_era5_1degree import GCSERA5OneDegree



import logging
from typing import Optional
import numpy as np
import pandas as pd
import xarray as xr

logger = logging.getLogger("ufs2arco")

class Source:
    """
    Base class for all ensemble forecast-like datasets
    """

    # these will be set by the different analysis, forecast, ensemble_forecast classes
    sample_dims = tuple()
    base_dims = tuple()

    # fill these out per subclass
    static_vars = tuple()
    available_variables = tuple()
    available_levels = tuple()

    @property
    def rename(self) -> dict:
        """
        Use this to map to the dimensions in sample_dims + base_dims
        """
        return dict()

    @property
    def dynamic_vars(self) -> tuple:
        return tuple(x for x in self.variables if x not in self.static_vars)

    @property
    def name(self) -> str:
        return self.__class__.__name__

    @property
    def _level_sel_kwargs(self) -> dict:
        if self.use_nearest_levels:
            return {"method": "nearest"}
        else:
            return {}

    def __init__(
        self,
        variables: Optional[list | tuple] = None,
        levels: Optional[list | tuple] = None,
        use_nearest_levels: Optional[bool] = False,
    ) -> None:
        """
        Initialize the Source object.

        Args:
            variables (list, tuple, optional): variables to grab
            levels (list, tuple, optional): vertical levels to grab
            use_nearest_levels (bool, optional): if True, all level selection with
                ``xarray.Dataset.sel(level=levels, method="nearest")``
        """

        # check variable selection
        if variables is not None:
            unrecognized = []
            for v in variables:
                if v not in self.available_variables:
                    unrecognized.append(v)

            if len(unrecognized) > 0:
                raise NotImplementedError(f"{self.name}.__init__: the following variables are not recognized or not implemented: {unrecognized}")
            self.variables = variables
        else:
            self.variables = list(self.available_variables)

        # check level selection
        self.use_nearest_levels = use_nearest_levels
        if levels is not None:
            unrecognized = []
            for l in levels:
                if l not in self.available_levels:
                    unrecognized.append(l)

            if len(unrecognized) > 0 and not use_nearest_levels:
                msg = f"{self.name}.__init__: the following levels are not recognized or not implemented:\n{unrecognized}"
                msg += "\nSet option 'use_nearest_levels=True' if you want to use xarray.Dataset.sel(level=levels, method='nearest')"
                raise ValueError(msg)
            self.levels = levels

        else:
            self.levels = list(self.available_levels)

        logger.info(str(self))

    def __str__(self) -> str:
        title = f"Source: {self.name}"
        msg = f"\n{title}\n" + \
              "".join(["-" for _ in range(len(title))]) + "\n"
        attrslist = list(self.sample_dims) + ["variables", "levels", "use_nearest_levels"]
        for key in attrslist:
            msg += f"{key:<18s}: {getattr(self, key)}\n"
        return msg

    def add_full_extra_coords(self, xds: xr.Dataset) -> xr.Dataset:
        """
        An optional routine that builds extra coordinates if needed, see example in ensemble_forecast.py.
        This gets used for passive targets
        """
        return xds



import logging
from typing import Optional
from copy import deepcopy

import numpy as np
import pandas as pd
import xarray as xr
import zarr

from ufs2arco.sources import Source, AnalysisSource, EnsembleForecastSource
from ufs2arco.targets import Target

logger = logging.getLogger("ufs2arco")

class Anemoi(Target):
    """
    Store dataset ready for anemoi

    Expected output has dimensions
        ``("time", "variable", "ensemble", "cell")``

    Use the rename argument to modify any of these before they get packed in the anemoi dataset.
    This might be useful if you want to train a model with the same variables, but from different datasets,
    so they have different names originally.

    Assumptions:
        * For :class:`EnsembleForecastSource` and :class:`ForecastSource` datasets t0 from ForecastSource gets renamed to time, and fhr is silently dropped
        * :attr:`do_flatten_grid` = ``True``
        * resolution = None, I have no idea where this gets set in anemoi-datasets
        * just setting use_level_index = False for now, but eventually it would be nice to use this flag to switch between how vertical level suffixes are labeled
    """

    # these should probably be options
    do_flatten_grid = True
    resolution = None
    use_level_index = False
    allow_nans = True

    # these are basically properties
    base_dims = ("variable", "cell")
    always_open_static_vars = True

    @property
    def sample_dims(self):
        if self._has_member:
            return ("time", "ensemble")
        else:
            return ("time",)

    @property
    def datetime(self):
        if self._has_fhr:
            #TODO: this is the hack where I assume we're using fhr=0 always
            return self.source.t0 + pd.Timedelta(hours=self.source.fhr[0])
        else:
            return self.source.time

    @property
    def time(self):

        if self._has_fhr:
            return np.arange(len(self.source.t0))
        else:
            return np.arange(len(self.source.time))

    @property
    def ensemble(self):
        if self._has_member:
            return self.source.member
        else:
            return [0]

    @property
    def protected_rename(self) -> dict:
        protected_rename = {
            "latitude": "latitudes",
            "longitude": "longitudes",
        }
        if self._has_fhr:
            protected_rename["t0"] = "dates"
        else:
            protected_rename["time"] = "dates"

        if self._has_member:
            protected_rename["member"] = "ensemble"
        return protected_rename


    @property
    def dim_order(self):
        if self.do_flatten_grid:
            return ("time", "variable", "ensemble", "cell")
        else:
            return ("time", "variable", "ensemble", "latitudes", "longitudes")


    def __init__(
        self,
        source: Source,
        chunks: dict,
        store_path: str,
        rename: Optional[dict] = None,
        slices: Optional[dict] = None,
        forcings: Optional[tuple | list] = None,
    ) -> None:

        super().__init__(
            source=source,
            chunks=chunks,
            store_path=store_path,
            rename=rename,
            slices=slices,
            forcings=forcings,
        )


        # additional checks
        if self._has_fhr:
            assert len(self.source.fhr) == 1 and self.source.fhr[0] == 0, \
                f"{self.name}.__init__: Can only use this class with fhr=0, no multiple lead times"

        renamekeys = list(self.rename.keys())
        protected = list(self.protected_rename.keys()) + list(self.protected_rename.values())
        for key in renamekeys:
            if key in protected or self.rename[key] in protected:
                logger.info(f"{self.name}.__init__: can't rename {key} -> {self.rename[key]}, either key or val is in a protected list. I'll drop it and forget about it.")
                self.rename.pop(key)

    def get_expanded_dim_order(self, xds):
        """this is used in :meth:`map_static_to_expanded`"""
        return ("time", "ensemble") + tuple(xds.attrs["stack_order"])


    def apply_transforms_to_sample(
        self,
        xds: xr.Dataset,
    ) -> xr.Dataset:

        if self._has_fhr:
            xds = xds.squeeze("fhr", drop=True)

        if not self._has_member:
            xds = xds.expand_dims({"ensemble": self.ensemble})

        xds = super().apply_transforms_to_sample(xds)
        xds = self._map_datetime_to_index(xds)
        xds = self._map_levels_to_suffixes(xds)
        xds = self._map_static_to_expanded(xds)
        xds = xds.transpose(*self.get_expanded_dim_order(xds))
        xds = self._stackit(xds)
        xds = self._calc_sample_stats(xds)
        if self.do_flatten_grid:
            xds = self._flatten_grid(xds)
        xds = xds.transpose(*self.dim_order)
        xds = xds.reset_coords()
        xds = xds[sorted(xds.data_vars)]
        return xds


    def manage_coords(self, xds: xr.Dataset) -> xr.Dataset:
        attrs = {
            "allow_nans": self.allow_nans,
            "ensemble_dimension": len(self.ensemble),
            "flatten_grid": self.do_flatten_grid,
            "resolution": str(self.resolution),
            "start_date": str(self.datetime[0]).replace(" ", "T"),
            "end_date": str(self.datetime[-1]).replace(" ", "T"),
            "frequency": self.datetime.freqstr,
            "statistics_start_date": str(self.datetime[0]).replace(" ", "T"),
            "statistics_end_date": str(self.datetime[-1]).replace(" ", "T"),
        }
        xds.attrs.update(attrs)
        return xds


    def rename_dataset(self, xds: xr.Dataset) -> xr.Dataset:
        """
        In addition to any user specified renamings...
        This takes the default source dimensions and renames them to the default anemoi dimensions:

        (t0, member, level, latitude, longitude) -> (dates, ensemble, level, latitudes, longitudes)
        or
        (time, level, latitude, longitude) -> (dates, ensemble, level, latitudes, longitudes)

        Args:
            xds (xr.Dataset): a dataset directly from the source

        Returns:
            xds (xr.Dataset): with renaming as above
        """
        # first, rename the protected list
        xds = xds.rename(self.protected_rename)
        xds = super().rename_dataset(xds)
        return xds


    def _map_datetime_to_index(self, xds: xr.Dataset) -> xr.Dataset:
        """
        Turn the datetime vector into a logical index and swap the dimensions

        (dates, ensemble, level, latitudes, longitudes) -> (time, ensemble, level, latitudes, longitudes)

        Args:
            xds (xr.Dataset): with time dimension "dates"

        Returns:
            xds (xr.Dataset): with new time dimension "time" ("dates" is still there)
        """

        t = [list(self.datetime).index(date) for date in xds["dates"].values]
        xds["time"] = xr.DataArray(
            t,
            coords=xds["dates"].coords,
            dims=xds["dates"].dims,
            attrs={
                "description": "logical time index",
            },
        )
        xds = xds.swap_dims({"dates": "time"})
        return xds

    def _map_levels_to_suffixes(self, xds):
        """
        Take each of the 3D variables, and make them n_levels x 2D variables with the suffix _{level} for each level

        (time, ensemble, level, latitudes, longitudes) -> (time, ensemble, latitudes, longitudes)

        Args:
            xds (xr.Dataset): with all variables, maybe 2D maybe 3D

        Returns:
            nds (xr.Dataset): 3D variables expanded into n_level x 2D variables
        """

        nds = xr.Dataset()
        nds.attrs["variables_metadata"] = dict()

        for name in xds.data_vars:
            meta = {
                "mars": {
                    "date": str(self.datetime[xds.time.values[0]]).replace("-","")[:8],
                    "param": name,
                    "step": 0, # this is the fhr=0 assumption
                    "time": str(self.datetime[xds.time.values[0]]).replace("-","").replace(" ", "").replace(":","")[8:12], # no idea what this should be actually
                    "valid_datetime": str(self.datetime[xds.time.values[0]]).replace(" ", "T"),
                    "variable": name,
                },
            }
            if len(self.ensemble) > 1:
                meta["mars"]["number"] = list(self.ensemble)
            if "level" in xds[name].dims:
                for level in xds[name].level.values:
                    idx = self._get_level_index(xds, level)
                    ilevel = int(level)
                    ilevel = ilevel if ilevel == level else level
                    suffix_name = f"{name}_{ilevel}" if not self.use_level_index else f"{name}_{idx}"
                    nds[suffix_name] = xds[name].sel({"level": level}, drop=True)
                    units = xds["level"].attrs.get('units', '')
                    nds[suffix_name].attrs.update(
                        {
                            "level": ilevel,
                            "level_description": f"{name} at vertical level (index, value) = ({idx}, {ilevel}{units})",
                            "level_index": idx,
                        },
                    )
                    nds.attrs["variables_metadata"][suffix_name] = deepcopy(meta)
                    nds.attrs["variables_metadata"][suffix_name]["mars"]["level"] = ilevel if not self.use_level_index else idx

                if "remapping" not in nds.attrs:
                    nds.attrs["remapping"] = {"param_level": "{param}_{levelist}"}
                else:
                    if "param_level" not in nds.attrs["remapping"].keys():
                        nds.attrs["remapping"]["param_level"] = "{param}_{levelist}"
            else:
                nds[name] = xds[name]

                if "computed_forcing" in xds[name].attrs and "constant_in_time" in xds[name].attrs:
                    nds.attrs["variables_metadata"][name] = {
                        "computed_forcing": xds[name].attrs["computed_forcing"],
                        "constant_in_time": xds[name].attrs["constant_in_time"],
                    }
                else:
                    nds.attrs["variables_metadata"][name] = deepcopy(meta)
                # Is attributes here a hack? Add the "field_shape" here
                # so that it's in the order of the data arrays, not in the dataset order
                # (they could be different)
                if "field_shape" not in nds.attrs:
                    nds.attrs["stack_order"] = list(d for d in xds[name].dims if d in ("latitudes", "longitudes"))
                    nds.attrs["field_shape"] = list(len(xds[d]) for d in nds.attrs["stack_order"])

        return nds

    @staticmethod
    def _get_level_index(xds: xr.Dataset, value: int | float) -> int | float:
        return xds["level"].values.tolist().index(value)

    def _map_static_to_expanded(self, xds: xr.Dataset) -> xr.Dataset:
        """
        Take each variable that does not have any of the (time, ensemble, latitudes, longitudes),
        and expand it so they all look the same

        Args:
            xds (xr.Dataset): with some static variable maybe

        Returns:
            xds (xr.Dataset): all data_vars have the same shape
        """

        for key in xds.data_vars:
            for d in xds.dims:
                if d not in xds[key].dims:
                    xds[key] = xds[key].expand_dims({d: xds[d]})

        return xds

    def _stackit(self, xds: xr.Dataset) -> xr.Dataset:
        """
        Stack the multivariate dataset to a single data array with all variables (and vertical levels) stacked together

        (time, ensemble, latitudes, longitudes) -> (time, ensemble, variable, latitudes, longitudes)

        Args:
            xds (xr.Dataset): with all variables, 3D variables as multiple 2D variables with suffixes

        Returns:
            xds (xr.Dataset): with "data" DataArray, which has all variables/levels stacked together
        """
        varlist = sorted(list(xds.data_vars))
        channel = [i for i, _ in enumerate(varlist)]
        channel = xr.DataArray(
            channel,
            coords={"variable": channel},
            dims="variable",
        )

        # this might be nice, but it doesn't exist in anemoi
        # and it causes problems with the container / fill workflow
        # it should just be added as a coordinate... but again it's not in anemoi
        #channel_names = xr.DataArray(
        #    varlist,
        #    coords=channel.coords,
        #    dims=channel.dims,
        #)

        data_vars = xr.concat(
            [
                xds[name].expand_dims({"variable": [this_channel]})
                for this_channel, name in zip(channel, varlist)
            ],
            dim="variable",
            combine_attrs="drop",
        )
        nds = data_vars.to_dataset(name="data")
        nds.attrs = xds.attrs.copy()
        # not making this a data array, even though it might be kinda nice
        nds.attrs["variables"] = varlist
        return nds

    def _flatten_grid(self, xds: xr.Dataset) -> xr.Dataset:
        """
        Flatten (latitudes, longitudes) -> (cell,)

        (time, ensemble, variable, latitudes, longitudes) -> (time, ensemble, variable, cell)

        Args:
            xds (xr.Dataset): with expanded grid

        Returns:
            xds (xr.Dataset): with grid flattened to "cell"
        """
        nds = xds.stack(cell2d=xds.attrs["stack_order"])
        nds["cell"] = xr.DataArray(
            np.arange(len(nds["cell2d"])),
            coords=nds["cell2d"].coords,
            dims=nds["cell2d"].dims,
            attrs={
                "description": f"logical index for 'cell2d', which is a flattened lon x lat array",
            },
        )
        nds = nds.swap_dims({"cell2d": "cell"})

        # For some reason, there's a failure when trying to store this multi-index
        # it's not needed in Anemoi, so no need to keep it anyway.
        nds = nds.drop_vars("cell2d")
        return nds

    def _calc_sample_stats(self, xds: xr.Dataset) -> xr.Dataset:
        """
        Compute statistics for this data sample, which will be aggregated later

        Args:
            xds (xr.Dataset): with just the data and coordinates

        Returns:
            xds (xr.Dataset): with the following statistics, each with an "_array" suffix,
                in order to indicate that the result will still have "time" and "ensemble" dimensions
                that will need to get aggregated
                ["count", "has_nans", "maximum", "minimum", "squares", "sums"]
        """

        dims = ["latitudes", "longitudes"]
        xds["count_array"] = (~np.isnan(xds["data"])).sum(dims, skipna=self.allow_nans).astype(np.int64)
        xds["has_nans_array"] = np.isnan(xds["data"]).any(dims)
        xds["maximum_array"] = xds["data"].max(dims, skipna=self.allow_nans).astype(np.float64)
        xds["minimum_array"] = xds["data"].min(dims, skipna=self.allow_nans).astype(np.float64)
        xds["squares_array"] = (xds["data"]**2).sum(dims, skipna=self.allow_nans).astype(np.float64)
        xds["sums_array"] = xds["data"].sum(dims, skipna=self.allow_nans).astype(np.float64)
        return xds

    def aggregate_stats(self) -> None:
        """Aggregate statistics over "time" and "ensemble" dimension...
        I'm assuming that this is relatively inexpensive without the spatial dimension

        This will store an array with the statistics

            ["count", "has_nans", "maximum", "mean", "minimum", "squares", "stdev", "sums"]

        and it will get rid of the "_array" versions of the statistics
        """

        xds = xr.open_zarr(self.store_path)

        # the easy ones
        xds["count"] = xds["count_array"].sum(["time", "ensemble"])
        xds["has_nans"] = xds["has_nans_array"].any(["time", "ensemble"])
        xds["maximum"] = xds["maximum_array"].max(["time", "ensemble"])
        xds["minimum"] = xds["minimum_array"].min(["time", "ensemble"])
        xds["squares"] = xds["squares_array"].sum(["time", "ensemble"])
        xds["sums"] = xds["sums_array"].sum(["time", "ensemble"])

        # now add mean & stdev
        xds["mean"] = xds["sums"] / xds["count"]
        variance = xds["squares"] / xds["count"] - xds["mean"]**2
        xds["stdev"] = xr.where(variance >= 0, np.sqrt(variance), 0.)

        # store it
        xds.to_zarr(self.store_path, mode="a")

        # now remove the temp versions
        zds = zarr.open(self.store_path, mode="a")
        for key in [
            "count_array",
            "has_nans_array",
            "maximum_array",
            "minimum_array",
            "squares_array",
            "sums_array",
        ]:
            logger.info(f"{self.name}.aggregate_statistics: Removing temporary version {key}")
            del zds[key]
        zarr.consolidate_metadata(self.store_path)



"""Tools for working with Gaussian grids."""
from __future__ import absolute_import, division, print_function

import numpy as np
import numpy.linalg as la
from numpy.polynomial.legendre import legcompanion, legder, legval

def gaussian_latitudes(n):
    """Construct latitudes and latitude bounds for a Gaussian grid.

    Args:
        n (int): The Gaussian grid number (half the number of latitudes in the grid.
    Returns:
        latitudes (numpy.array): length ``n`` array of latitudes in degrees
        bounds2d (numpy.array): ``(n, 2)`` array of grid bounds
    """
    if abs(int(n)) != n:
        raise ValueError("n must be a non-negative integer")
    nlat = 2 * n
    # Create the coefficients of the Legendre polynomial and construct the
    # companion matrix:
    cs = np.array([0] * nlat + [1], dtype=int)
    cm = legcompanion(cs)
    # Compute the eigenvalues of the companion matrix (the roots of the
    # Legendre polynomial) taking advantage of the fact that the matrix is
    # symmetric:
    roots = la.eigvalsh(cm)
    roots.sort()
    # Improve the roots by one application of Newton's method, using the
    # solved root as the initial guess:
    fx = legval(roots, cs)
    fpx = legval(roots, legder(cs))
    roots -= fx / fpx
    # The roots should exhibit symmetry, but with a sign change, so make sure
    # this is the case:
    roots = (roots - roots[::-1]) / 2.0
    # Compute the Gaussian weights for each interval:
    fm = legval(roots, cs[1:])
    fm /= np.abs(fm).max()
    fpx /= np.abs(fpx).max()
    weights = 1.0 / (fm * fpx)
    # Weights should be symmetric and sum to two (unit weighting over the
    # interval [-1, 1]):
    weights = (weights + weights[::-1]) / 2.0
    weights *= 2.0 / weights.sum()
    # Calculate the bounds from the weights, still on the interval [-1, 1]:
    bounds1d = np.empty([nlat + 1])
    bounds1d[0] = -1
    bounds1d[1:-1] = -1 + weights[:-1].cumsum()
    bounds1d[-1] = 1
    # Convert the bounds to degrees of latitude on [-90, 90]:
    bounds1d = np.rad2deg(np.arcsin(bounds1d))
    bounds2d = np.empty([nlat, 2])
    bounds2d[:, 0] = bounds1d[:-1]
    bounds2d[:, 1] = bounds1d[1:]
    # Convert the roots from the interval [-1, 1] to latitude values on the
    # interval [-90, 90] degrees:
    latitudes = np.rad2deg(np.arcsin(roots))
    return latitudes, bounds2d



import logging
from typing import Optional

import numpy as np
import xarray as xr

try:
    import flox
    _has_flox = True
except ImportError:
    _has_flox = False

logger = logging.getLogger("ufs2arco")

def fv_vertical_regrid(
    xds: xr.Dataset,
    weight_var: str,
    interfaces: list | np.ndarray,
    use_nearest_interfaces: Optional[bool] = False,
    keep_weight_var: Optional[bool] = False,
) -> xr.Dataset:
    """Vertically regrid a dataset based on interface values

    Args:
        xds (xr.Dataset)
        weight_var (str): with the name of the variable to weight regridding by
        interfaces (array_like): array of interface values to average vertical coordinate "level" between
            if "interface" does not exist in the dataset, this will be added, otherwise we subsample
        use_nearest_interfaces (bool, optional): if True, and "interface" is in the dataset,
            grab interfaces from dataset via
            ``xarray.Dataset.sel(interface=interfaces, method='nearest')``
        keep_weight_var (bool, optional): if False, drop the weight variable from the dataset before saving

    Returns:
        xds (xr.Dataset): with vertical averaging
    """

    if not _has_flox:
        logger.warning("Could not import flox, install with 'conda install -c conda-forge flox' for faster volume averaging (i.e. groupby operations)")

    assert weight_var in xds, \
        f"fv_vertical_regrid: can't find {weight_var} in dataset, can't use it for regridding"

    # get or create the exact interface values
    if "interface" in xds:
        kw = {"method": "nearest"} if use_nearest_interfaces else {}
        try:
            xds = xds.sel(interface=interfaces, **kw)
        except ValueError:
            msg = f"fv_vertical_regrid: couldn't find specified interfaces"
            msg += "\nSet option 'use_nearest_interfaces=True' if you want to use xarray.Dataset.sel(interface=interfaces, method='nearest')"
            raise ValueError(msg)

    else:
        xds["interface"] = xr.DataArray(
            interfaces,
            coords={"interface": interfaces},
            dims="interface",
            attrs={"description": "vertical coordinate interfaces created for regridding"},
        )

    # get the regrid weighting
    layer_thickness = xds[weight_var].groupby_bins(
        "level",
        bins=xds["interface"],
    ).sum()
    layer_thickness.attrs = xds[weight_var].attrs.copy()
    layer_thickness_inverse = 1/layer_thickness

    # do the regridding
    vars3d = [x for x in xds.data_vars if "level" in xds[x].dims and x != weight_var]
    for key in vars3d:
        attrs = xds[key].attrs.copy()
        xds[key] = layer_thickness_inverse * (
            (
                xds[key]*xds[weight_var]
            ).groupby_bins(
                "level",
                bins=xds["interface"],
            ).sum()
        )
        xds[key].attrs = attrs
        long_name = attrs.get("long_name", key)
        xds[key].attrs["long_name"] = f"vertically regridded {long_name}"

    # handle the bins and add attrs
    xds["level_bins"] = xds["level_bins"].swap_dims({"level_bins": "level"})
    for key in vars3d:
        with xr.set_options(keep_attrs=True):
            xds[key] = xds[key].swap_dims({"level_bins": "level"})
        xds[key].attrs["vertical_coordinate"] = f"{weight_var} weighted average in vertical, new coordinate bounds represented by 'interface'"

    if keep_weight_var:
        xds[weight_var] = layer_thickness.swap_dims({"level_bins": "level"})
        xds[weight_var].attrs["vertical_coordinate"] = f"vertically averaged, new coordinate bounds represented by 'interface'"
    else:
        xds = xds.drop_vars(weight_var)

    # make new coordinates for approximate new level
    new_level = (xds["interface"].values[:-1] + xds["interface"].values[1:])/2
    if np.all(new_level == new_level.astype(int)):
        new_level = new_level.astype(int)
    xds["new_level"] = xr.DataArray(
        new_level,
        coords=xds["level"].coords,
        attrs={
            "description": f"approximated vertical grid cell center after regridding",
            "details": "computed as (interface[:-1] + interface[1:])/2",
        },
    )
    xds = xds.swap_dims({"level": "new_level"})
    xds = xds.drop_vars("level")
    xds = xds.rename({"new_level": "level"})

    # unfortunately, cannot store the level_bins due to this issue: https://github.com/pydata/xarray/issues/2847
    xds = xds.drop_vars("level_bins")
    return xds



import logging
from typing import Optional

import pandas as pd
import xarray as xr

from ufs2arco.sources import AnalysisSource

logger = logging.getLogger("ufs2arco")


class GCSERA5OneDegree(AnalysisSource):

    static_vars = ("land_sea_mask", "geopotential_at_surface")

    _is_open = False

    @property
    def available_variables(self) -> tuple:
        return (
            "10m_u_component_of_wind",
            "10m_v_component_of_wind",
            "2m_temperature",
            "angle_of_sub_gridscale_orography",
            "anisotropy_of_sub_gridscale_orography",
            "geopotential",
            "geopotential_at_surface",
            "high_vegetation_cover",
            "lake_cover",
            "lake_depth",
            "land_sea_mask",
            "low_vegetation_cover",
            "mean_sea_level_pressure",
            "sea_ice_cover",
            "sea_surface_temperature",
            "slope_of_sub_gridscale_orography",
            "soil_type",
            "specific_humidity",
            "standard_deviation_of_filtered_subgrid_orography",
            "standard_deviation_of_orography",
            "surface_pressure",
            "temperature",
            "toa_incident_solar_radiation",
            "total_cloud_cover",
            "total_column_water_vapour",
            "total_precipitation",
            "type_of_high_vegetation",
            "type_of_low_vegetation",
            "u_component_of_wind",
            "v_component_of_wind",
            "vertical_velocity",
        )

    @property
    def available_levels(self) -> tuple:
        return (
            1, 2, 3, 5, 7, 10, 20, 30, 50, 70,
            100, 125, 150, 175, 200, 225, 250,
            300, 350, 400, 450, 500, 550,
            600, 650, 700, 750, 775,
            800, 825, 850, 875,
            900, 925, 950, 975, 1000,
        )

    def _open_dataset(self):
        """May as well just do all of this once
        """

        # open
        xds = xr.open_zarr(
            "gs://gcp-public-data-arco-era5/ar/1959-2022-1h-360x181_equiangular_with_poles_conservative.zarr",
            storage_options={"token": "anon"},
        )

        # select
        xds = xds[self.variables]
        xds = xds.sel(level=self.levels, **self._level_sel_kwargs)

        # set
        self._xds = xds
        self._is_open = True

    def open_sample_dataset(
        self,
        time: pd.Timestamp,
        open_static_vars: bool,
        cache_dir: Optional[str] = None,
    ) -> xr.Dataset:

        if not self._is_open:
            self._open_dataset()

        xds = self._xds.sel(time=[time])
        selection = list(self.variables) if open_static_vars else list(self.dynamic_vars)
        xds = xds[selection].copy().load()
        return xds



import logging

import xarray as xr

from ufs2arco.transforms.vertical_regrid import fv_vertical_regrid
from ufs2arco.transforms.mappings import get_available_mappings, apply_mappings

logger = logging.getLogger("ufs2arco")

class Transformer:

    @property
    def implemented(self) -> tuple:
        return (
            "fv_vertical_regrid",
            "mapping",
        )

    def __init__(self, options):

        names = list(options.keys())

        unrecognized = []

        # first check regrid, mapping
        for name in names:
            if name not in self.implemented:
                unrecognized.append(name)

        # now check for mappings
        if "mapping" in names:
            available = list(get_available_mappings().keys())
            for mapname in options["mapping"]:
                if mapname not in available:
                    unrecognized.append(f"mapping: {mapname}")

        if len(unrecognized) > 0:
            raise NotImplementedError(f"Transformer.__init__: the following transformations are not recognized or not implemented: {unrecognized}")

        self.names = names
        self.options = options

    def __call__(self, xds: xr.Dataset):
        """
        Process the dataset, performing any of the desired transformations

        Args:
            xds (xr.Dataset): with the data

        Returns:
            xds (xr.Dataset): the processed version
        """

        if "fv_vertical_regrid" in self.names:
            xds = fv_vertical_regrid(xds, **self.options["fv_vertical_regrid"])

        if "mapping" in self.names:
            xds = apply_mappings(xds, self.options["mapping"])

        return xds



__version__ = "0.4.2"

from .cice6dataset import CICE6Dataset
from .fv3dataset import FV3Dataset
from .layers2pressure import Layers2Pressure
from .mom6dataset import MOM6Dataset
from .regrid import MOM6Regridder, CICE6Regridder
from .timer import Timer



import logging
import sys
import numpy as np
import dask
import xarray as xr
import xesmf
import cf_xarray as cfxr

from ufs2arco.regrid.gaussian_grid import gaussian_latitudes
from ufs2arco import Layers2Pressure

class SimpleFormatter(logging.Formatter):
    def format(self, record):
        record.relativeCreated = record.relativeCreated // 1000
        return super().format(record)

def setup_log(level=logging.INFO):

    logging.basicConfig(
        stream=sys.stdout,
        level=level,
    )
    logger = logging.getLogger()
    formatter = SimpleFormatter(fmt="[%(relativeCreated)d s] [%(levelname)s] %(message)s")
    for handler in logger.handlers:
        handler.setFormatter(formatter)

def open_datasets():
    """
    Returns:
        predictions, targets, original_target_dataset, independent_truth_dataset
    """

    rds = xr.open_zarr(
        "gs://noaa-ufs-gefsv13replay/ufs-hr1/0.25-degree/03h-freq/zarr/fv3.zarr",
        storage_options={"anon":True},
    )
    rds = rds.drop_vars(["cftime", "ftime"])
    rds = rds.where(rds.time.dt.hour / 6 == rds.time.dt.hour // 6, drop=True)
    rds = rds.isel(time=np.random.randint(low=0,high=len(rds.time), size=20))

    rds = rds[["pressfc", "hgtsfc", "tmp", "spfh", "delz"]]
    logging.info(f"Loading {rds.nbytes / 1e9} GB dataset...")
    rds = rds.load()
    logging.info(" ... done")
    rds = rds.sortby("time")

    truth = xr.open_zarr(
        "gs://weatherbench2/datasets/era5/1959-2023_01_10-6h-240x121_equiangular_with_poles_conservative.zarr",
        storage_options={"token":"anon"},
    )

    # subsample in space to avoid poles
    truth = truth.sel(latitude=slice(-89, 89))

    rds = rds.rename({"pfull": "level", "grid_xt": "lon", "grid_yt": "lat"})
    return rds, truth


def get_bounds(xds, is_gaussian=False):
    xds = xds.cf.add_bounds(["lat", "lon"])

    for key in ["lat", "lon"]:
        corners = cfxr.bounds_to_vertices(
            bounds=xds[f"{key}_bounds"],
            bounds_dim="bounds",
            order=None,
        )
        xds = xds.assign_coords({f"{key}_b": corners})
        xds = xds.drop_vars(f"{key}_bounds")

    if is_gaussian:
        xds = xds.drop_vars("lat_b")
        _, lat_b = gaussian_latitudes(len(xds.lat)//2)
        lat_b = np.concatenate([lat_b[:,0], [lat_b[-1,-1]]])
        if xds["lat"][0] > 0:
            lat_b = lat_b[::-1]
        xds["lat_b"] = xr.DataArray(
            lat_b,
            dims="lat_vertices",
        )
        xds = xds.set_coords("lat_b")
    return xds

def create_output_dataset(lat, lon, is_gaussian):
    xds = xr.Dataset({
        "lat": lat,
        "lon": lon,
    })
    return get_bounds(xds, is_gaussian)


def regrid_and_rename(xds, truth):
    """Note that it's assumed the truth dataset is not on a Gaussian grid but input is"""

    ds_out = create_output_dataset(
        lat=truth["latitude"].values,
        lon=truth["longitude"].values,
        is_gaussian=False,
    )
    if "lat_b" not in xds and "lon_b" not in xds:
        xds = get_bounds(xds, is_gaussian=True)

    regridder = xesmf.Regridder(
        ds_in=xds,
        ds_out=ds_out,
        method="conservative",
        reuse_weights=False,
    )
    ds_out = regridder(xds, keep_attrs=True)

    rename_dict = {
        "pressfc": "surface_pressure",
        "ugrd10m": "10m_u_component_of_wind",
        "vgrd10m": "10m_v_component_of_wind",
        "tmp2m": "2m_temperature",
        "tmp": "temperature",
        "ugrd": "u_component_of_wind",
        "vgrd": "v_component_of_wind",
        "dzdt": "vertical_velocity",
        "spfh": "specific_humidity",
        "prateb_ave": "total_precipitation_3hr",
        "lat": "latitude",
        "lon": "longitude",
    }
    rename_dict = {k: v for k,v in rename_dict.items() if k in ds_out}
    ds_out = ds_out.rename(rename_dict)

    # ds_out has the lat/lon boundaries from input dataset
    # remove these because it doesn't make sense anymore
    ds_out = ds_out.drop_vars(["lat_b", "lon_b"])
    return ds_out


def interp2pressure(xds, plevels, hgtsfc=None):
    """Assume plevels is in hPa"""

    lp = Layers2Pressure()
    prsl = lp.calc_layer_mean_pressure(xds["pressfc"], xds["tmp"], xds["spfh"], xds["delz"])

    if "geopotential" not in xds and hgtsfc is not None:
        xds["geopotential"] = lp.calc_geopotential(hgtsfc, xds["delz"])
        xds["geopotential"] = xds["geopotential"].chunk({
            "time": 1,
            "level": 1,
            "lat": -1,
            "lon": -1,
        })

    vars3d = ["geopotential"]
    pds = xr.Dataset()
    plevels = np.array(list(plevels))
    pds["level"] = xr.DataArray(
        plevels,
        coords={"level": plevels},
        dims=("level",),
        attrs={
            "description": "Pressure level",
            "units": "hPa",
        },
    )
    results = {k: list() for k in vars3d}
    for p in plevels:


        cds = lp.get_interp_coefficients(p*100, prsl)
        mask = (cds["is_right"].sum("level") > 0) & (cds["is_left"].sum("level") > 0)
        for key in vars3d:
            logging.info(f"Interpolating {key} to {p} hPa")
            interpolated = lp.interp2pressure(xds[key], p*100, prsl, cds)
            interpolated = interpolated.expand_dims({"level": [p]})
            interpolated = interpolated.where(mask)
            results[key].append(interpolated.compute())

    for key in vars3d:
        pds[key] = xr.concat(results[key], dim="level")

    return pds


if __name__ == "__main__":

    setup_log()
    dask.config.set(scheduler="threads", num_workers=48)

    rds, truth = open_datasets()
    hgtsfc = rds["hgtsfc"].isel(time=0).drop_vars(["time"])
    hgtsfc = hgtsfc.load()
    rds["hgtsfc"] = hgtsfc

    rds = interp2pressure(rds, [100, 500, 850], hgtsfc)
    logging.info(f"Interpolated to pressure levels...")

    # regrid and rename variables
    rds = regrid_and_rename(rds, truth)
    logging.info(f"Done regridding...")

    path = f"/p1-evaluation/v1/validation/replay.geopotential.20.zarr"
    rds.to_zarr(path)
    logging.info(f"Done writing to {path}")
